const LogHelper                = require("../helpers/LogHelper.js");
const config                   = require("../config/api.js");
const routes                   = require("../config/routes.js");

const fs                       = require("fs");
const https                    = require("https");
const http                     = require("http");

const express                  = require("express");
const bodyParser               = require("body-parser");
const helmet                   = require("helmet");

const rootDirectory = "./src";

/**
 * Represents the Api
 */
class Api {

    /**
     * Initializes the Api components
     * @static
     * @private
     * @param {String} directory - Path to the component code
     * @return {Object} Object tree of initialized components
     */
    static _buildObjectByFiles(directory){
        const obj = {};
        const files = fs.readdirSync(rootDirectory + "/" + directory);
        files.forEach((file) => {
            const name = file.replace('.js', '');
            if (directory !== "config" && name.indexOf("Abstract") === -1){
                const classInstance = require(`../${directory}/${name}.js`);
                obj[name] = new classInstance();
            }
        });
        return obj;
    }

    /**
     * Linking the route and HTTP Method to its controller action
     * @static
     * @private
     * @param {Express} app - Express application
     * @param {Array} routes - Array of route definitions
     * @param {Array} controllers - Array of controller instances
     */
    static _buildRoutes(app, routes, controllers){
        for (let key in routes){
            const method = key.replace(/\/.*$/, '').trim();
            const path = key.replace(/^[a-z ]*/, '').trim();

            const requestHandles = routes[key].split('|');
            const controllerPair = requestHandles[0].trim().split(".");
            const name = controllerPair[0];
            const handle = controllerPair[1];

            if (requestHandles.length === 2){
                const policyName = requestHandles[1].trim();
                const policyClass = require(`../policies/${policyName}.js`);
                app[method](path, policyClass.validateRequest, controllers[name][handle]);
            }else{
                app[method](path, controllers[name][handle]);
            }
        }
    }

    /**
     * Add express middleware and set routing
     * @static
     * @private
     * @param {Api} api - Api instance
     */
    static _initialize(api){
        api.app.use(helmet());
        api.app.use(bodyParser.json());
        api.app.use((err, req, res, next) => {
            if (err)
                res.status(400).send();
            else
                next();
        });

        const includes = { controllers: {} } ;
        for (let key in includes){
            includes[key] = Api._buildObjectByFiles(key);
        }

        Api._buildRoutes(api.app, routes, includes.controllers);
        api._log().info("Api is ready");
    }

    /**
     * Create Api instance
     */
    constructor() {
        this.app = express();
    }

    /**
     * Start the api server
     * @returns {Promise} Resolve if server is started
     */
    async start(){
        this._initializationServices();
        Api._initialize(this);
        try {
            let server = null;
            if (config.enableHTTPS){
                https.createServer({
                    key: await fs.promises.readFile(config.certificate.key),
                    cert: await fs.promises.readFile(config.certificate.cert),
                    ca: await fs.promises.readFile(config.certificate.ca)
                }, this.app).listen(config.port, "localhost");
            }else{
                server = http.createServer(this.app).listen(config.port, 'localhost');
            }
            this._log().info("Staring server on port " + config.port);
        } catch(err){
            this._log().error("Start of server failed: " + err);
        }
    }

    /**
     * Get app object
     * @returns {Express} Express object
     */
    getApp(){
        return this.app;
    }

    /**
     * Load all services
     * @private
     */
    _initializationServices(){
        const serviceDirectories = fs.readdirSync(`${rootDirectory}/services`);
        serviceDirectories.forEach(d => {
            if (fs.existsSync(`${rootDirectory}/services/${d}/${d}Service.js`)){
                const ServiceClass = require(`../services/${d}/${d}Service`);
                ServiceClass.initialize();
                this._log().info(`Service: ${d} is initialized`);
            }
        });
    }

    /**
     * Get logger
     * @private
     * @return {Logger} Logger instance
     */
    _log(){
        return LogHelper.getInstance().getLogger();
    }
}

module.exports = Api;
