const AbstractController        = require("./AbstractController.js");
const LernstickConnectorService = require("../services/LernstickConnector/LernstickConnectorService.js")

/**
 * Manages examination requests
 *  @extends AbstractController
 */
class ExaminationController extends AbstractController {

    /**
     * Create a ExaminationController
     */
    constructor() {
        super();

        this.lernstickservice = LernstickConnectorService.getInstance();

        this.create     = this.create.bind(this);
        this.start      = this.start.bind(this);
        this.stop       = this.stop.bind(this);
        this.getStatus  = this.getStatus.bind(this);
    }


    /**
     * @openapi
     * /examination:
     *   post:
     *     tags:
     *      - ExaminationController
     *     description: Create a new examination
     *     requestBody:
     *       required: true
     *       content:
     *         application/json:
     *           schema:
     *             type: object
     *             properties:
     *               examinationFile:
     *                  type: string
     *                  description: Base64 encoded ZIP archive with
     *     responses:
     *       '201':
     *          description: Created examination
     *       '412':
     *          $ref: '#/components/responses/PreconditionFailed'
     *       '500':
     *          $ref: '#/components/responses/ServerError'
     */
    async create(req, res) {
        try {
            if (typeof req.body.examinationFile === "undefined"){
                res.sendStatus(412);
                return;
            }

            const examinationConfig = await this.lernstickservice.verifyExaminationFile(req.body);
            if (examinationConfig === null){
                res.sendStatus(412);
                return;
            }

            const status = await this.lernstickservice.saveExaminationFile(req.body, examinationConfig);
            if (status){
                res.status(201);
                res.send({status: status});
            }else{
                res.sendStatus(412);
            }
        }catch (err){
            this._log().error(err);
            res.sendStatus(500);
        }
    }

    /**
     * @openapi
     * /examination/start:
     *   put:
     *     tags:
     *      - ExaminationsController
     *     description: Start examination
     *     responses:
     *       '200':
     *          description: Start examination
     *       '500':
     *          $ref: '#/components/responses/ServerError'
     */
    async start(req, res) {
        try {
            this.lernstickservice.startExamination();
            res.sendStatus(200);
        }catch (err){
            this._log().error(err);
            res.sendStatus(500);
        }
    }

    /**
     * @openapi
     * /examination/stop:
     *   put:
     *     tags:
     *      - ExaminationsController
     *     description: Start examination
     *     responses:
     *       '200':
     *          description: Start examination
     *       '500':
     *          $ref: '#/components/responses/ServerError'
     */
    async stop(req, res) {
        try {
            this.lernstickservice.stopExamination();
            res.sendStatus(200);
        }catch (err){
            this._log().error(err);
            res.sendStatus(500);
        }
    }


    /**
     * @openapi
     * /examination/status:
     *   get:
     *     tags:
     *      - ExaminationsController
     *     description: Get examination status
     *     responses:
     *       '200':
     *          description: Examination status
     *       '500':
     *          $ref: '#/components/responses/ServerError'
     */
    async getStatus(req, res) {
        try {
            const examination = this.lernstickservice.getExamination();
            res.status(200);
            res.send(examination);
        }catch (err){
            this._log().error(err);
            res.sendStatus(500);
        }
    }

}

module.exports = ExaminationController;
