const queue = require("queue");

/**
 * This is the definition of an event object. All events follow this layout
 * @typedef {Event} Event
 * @property {String}   action   - Possible actions are: create, update, remove
 * @property {String}   model    - Name of the model
 * @property {String}   type     - Type of the event normal, error
 * @property {Object}   object   - Payload of the event depending on the model
 */

/**
 * Handel all fired events from the models.
 */
class EventHelper {

    /**
     * Get the instance of the EventHelper
     * @static
     * @returns {EventHelper} EventHelper instance
     */
    static getInstance(){
        if (typeof EventHelper.instance === "undefined"){
            return new EventHelper();
        }else{
            return EventHelper.instance;
        }
    }

    /**
     * Singleton: Creates a EventHelper instance
     * @returns {EventHelper} Instance
     */
    constructor(){
        if (!!EventHelper.instance) {
            return EventHelper.instance;
        }
        EventHelper.instance = this;
        this.eventListeners  = [];
        this.eventQueue      = queue({ concurrency: 1 , autostart: true});
        return this;
    }

    /**
     * Add listener to event listeners
     * @param {Function} listener - Listener to catch get events
     */
    registerListener(listener){
        this.eventListeners.push(listener);
    }

    /**
     * Notifies all listeners
     * @param {String} action - Possible actions are: create, update, remove
     * @param {String} model - Name of the model
     * @param {Object} payload - Payload of the event depending on the model
     */
    fireEvent(action, model, payload){
        this.eventQueue.push(done => {
            const event = {
                action: action,
                model: model,
                type: "normal",
                object: payload
            };
            this.eventListeners.forEach(l => l(event));
            done();
        });
    }

}

module.exports = EventHelper;
