const config = require("../config/api.js");

const log4js = require("log4js");
const syslog = require("syslog-client");

const logSeverityMap = new Map();
logSeverityMap.set("trace", syslog.Severity.Debug);
logSeverityMap.set("debug", syslog.Severity.Debug);
logSeverityMap.set("info", syslog.Severity.Informational);
logSeverityMap.set("warn", syslog.Severity.Warning);
logSeverityMap.set("error", syslog.Severity.Error);
logSeverityMap.set("fatal", syslog.Severity.Critical);

/**
 * Manages the logging of the application
 */
class LogHelper {

    /**
     * Initializes an log helper instance
     * @static
     * @returns {LogHelper} Instance
     */
    static getInstance(){
        if (typeof LogHelper.instance === "undefined"){
            return new LogHelper();
        }else{
            return LogHelper.instance;
        }
    }

    /**
     * Singleton: Creates a log helper instance
     * @returns {LogHelper} Instance
     */
    constructor(){
        if (!!LogHelper.instance) {
            return LogHelper.instance;
        }
        LogHelper.instance = this;

        log4js.configure(config.log.log4js);
        const errorLogger  = log4js.getLogger("default");
        const auditLogger  = log4js.getLogger("audit");
        this.accessLogger = log4js.getLogger("access");


        if (config.log.syslog.enabled){
            config.log.syslog.options["transport"] = config.log.syslog.options["transport"].toLowerCase() === "tcp"
                ? syslog.Transport.Tcp : syslog.Transport.Udp;
            this.syslogClient = syslog.createClient(config.log.syslog.target, config.log.syslog.options);
            this.errorLogger  = this._logWrapper(errorLogger, "Error");
            this.auditLogger  = this._logWrapper(auditLogger, "Audit");
        }else{
            this.errorLogger = errorLogger;
            this.auditLogger = auditLogger;
        }

        return this;
    }

    /**
     * Get logger instance
     * @returns {Logger} Logger instance
     */
    getLogger(){
        return this.errorLogger;
    }

    /**
     * Get access logger instance
     * @returns {Logger} Access logger instance
     */
    getAccessLogger(){
        return log4js.connectLogger(this.accessLogger, { level: "info" });
    }

    /**
     * Write audit log
     * @param {Integer} role - User roles
     * @param {String} id - User / device identifier
     * @param {String} severity - Log severity
     * @param {String} msg - Log message
     */
    audit(role, id, severity, msg){
        this.auditLogger[severity](`[${role}:${id}] ${msg}`);
    }

    /**
     * Wraps the log4js logger and the syslog client logger
     * @param {Logger} logger - Log4js logger instance
     * @param {String} loggerName - Syslog log type
     * @private
     */
    _logWrapper(logger, loggerName){
        const wrapper = {};
        logSeverityMap.forEach((v, k) => {
            wrapper[k] = log => {
                logger[k](log);
                this.syslogClient.log(`${loggerName} ${log}`, { severity: v }, err => {
                    if (err) logger.error(err);
                });
            };
        });
        return wrapper;
    }
}

module.exports = LogHelper;
