const Service                   = require("../Service.js");
const EventHelper               = require("../../helpers/EventHelper.js");
const config                    = require("../../config/api.js");

const http                      = require("http");

const express                   = require("express");
const bodyParser                = require("body-parser");
const helmet                    = require("helmet");

/**
 * Events Service
 * @extends Service
 */
class EventsService extends Service {

    /**
     * Initializes the singleton
     * @static
     */
    static initialize() {
        if (typeof EventsService.instance === "undefined") {
            new EventsService();
        }
    }

    /**
     * Initializes a Events Service instance
     * @static
     * @returns {EventsService} Instance
     */
    static getInstance() {
        if (typeof EventsService.instance === "undefined") {
            return new EventsService();
        } else {
            return EventsService.instance;
        }
    }

    /**
     * Singleton: Creates a Events Service instance
     * @returns {EventsService} Instance
     */
    constructor() {
        super();
        if (!!EventsService.instance) {
            return EventsService.instance;
        }
        EventsService.instance      = this;
        this.app                    = express();
        this.connectionId           = 0;
        this.connections            = new Map();

        this.events                 = this.events.bind(this);

        this._initUserConnections   = this._initUserConnections.bind(this);
        this._closeConnection       = this._closeConnection.bind(this);
        this._sendEvent             = this._sendEvent.bind(this);
        this._notify                = this._notify.bind(this);

        this._start                 = this._start.bind(this);

        this._start();
        return this;
    }

    /**
     * Start Events server
     * @private
     */
    _start(){
        this.app.use(helmet());
        this.app.use(bodyParser.json());
        this.app.use((err, req, res, next) => {
            if (err)
                res.status(400).send();
            else
                next();
        });

        this.app["get"]("*", this.events);

        this._log().info("Staring events server on port " + config.services.Events.eventsPort);
        try {
            http.createServer(this.app).listen(config.services.Events.eventsPort, 'localhost');
            this._notify();
            this._log().info("Running events server on port " + config.services.Events.eventsPort);
        } catch(err){
            this._log().error("Start of events server failed: " + err);
        }
    }



    /**
     * @openapi
     * /:
     *   get:
     *     tags:
     *      - EventsController
     *     description: Handle requests to open an event stream
     *     responses:
     *       '200':
     *          description: Stream is starting
     *          headers:
     *              'Content-Type':
     *                  schema:
     *                      type: string
     *                      enum: ["text/event-stream"]
     *       '500':
     *          $ref: '#/components/responses/ServerError'
     */
    async events(req, res) {
        try {
            const connection = this._initUserConnections(req, res);

            // Assign close listeners
            req.on("close", () => this._closeConnection(connection));
            req.on("end",   () => this._closeConnection(connection));

            // Send initial message
            this._sendEvent(connection,{isConnected: true});
        }catch (err){
            this._log().warn(err);
            // Too many requests
            res.sendStatus(429);
        }
    }

    /**
     * Initialize user connection to get events
     * @param {Object} req - HTTP request
     * @param {Object} res - HTTP response
     * @private
     */
    _initUserConnections(req, res) {
        const id = this.connectionId + 1;
        req.socket.setTimeout(0);
        req.socket.setNoDelay(true);
        req.socket.setKeepAlive(true);
        res.statusCode = 200;
        res.setHeader('Content-Type', 'text/event-stream');
        res.setHeader('Cache-Control', 'no-cache');
        res.setHeader('X-Accel-Buffering', 'no');
        if (req.httpVersion !== '2.0') res.setHeader('Connection', 'keep-alive');
        this.connectionId = id;
        const connection = {req, res, id, msgId: 0};
        this.connections.set(id, connection);
        return connection;
    }



    /**
     * Remove a closed user connection
     * @param {Object} connection - Connections of a user
     * @private
     */
    _closeConnection(connection){
        if (this.connections.has(connection.id)) this.connections.delete(connection.id);
    }


    /**
     * Send event to consumers
     * @param {Object} connection - Connections of a user
     * @param {Event} event - Event to send
     * @private
     */
    _sendEvent(connection, event){
        const curr = this.connections.get(connection.id);
        const msgId = curr.msgId + 1;
        this.connections.set(connection.id, {
            req: curr.req,
            res: curr.res,
            id: connection.id,
            msgId: msgId
        });
        curr.res.write(`id: ${msgId}\n`);
        curr.res.write(`data: ${JSON.stringify(event)}\n\n`);
    }


    /**
     * Notify all clients of a user about occurring events
     * @private
     */
    _notify(){
        EventHelper.getInstance().registerListener((event) =>
            this.connections.forEach(c => this._sendEvent(c, event)));
    }


}

module.exports = EventsService;
