const Service                   = require("../Service.js");
const EventHelper               = require("../../helpers/EventHelper.js");
const config                    = require("../../config/api.js");

const { spawn }                 = require("child_process");
const fs                        = require("fs");
const path                      = require("path");
const crypto                    = require("crypto");

const streamZip                 = require("node-stream-zip");
const fetch                     = require("node-fetch")
const ajv                       = require("ajv/dist/2019");
const addFormats                = require("ajv-formats")
const uuid                      = require("uuid");

const ajvInstance               = new ajv();
addFormats(ajvInstance);

/**
 * Lernstick Connector Service
 * @extends Service
 */
class LernstickConnectorService extends Service {

    /**
     * Initializes the singleton
     * @static
     */
    static initialize() {
        if (typeof LernstickConnectorService.instance === "undefined") {
            new LernstickConnectorService();
        }
    }

    /**
     * Initializes a Lernstick Connector Service instance
     * @static
     * @returns {LernstickConnectorService} Instance
     */
    static getInstance() {
        if (typeof LernstickConnectorService.instance === "undefined") {
            return new LernstickConnectorService();
        } else {
            return LernstickConnectorService.instance;
        }
    }

    /**
     * Recursive remove of a directory
     * @param {String} delPath - Target path to remove
     * @return {Promise<boolean>}
     */
    static async removeDirectory(delPath){
        try {
            if (!fs.existsSync(delPath)) return true;
            const list = await fs.promises.readdir(delPath);
            for (let i = 0; i < list.length; i++) {
                const filepath = path.join(delPath, list[i]);
                const stat = fs.statSync(filepath);
                if (list[i] !== "." && list[i] !== "..") {
                    if (stat.isDirectory()) {
                        await LernstickConnectorService.removeDirectory(filepath);
                    }else{
                        await fs.promises.unlink(filepath);
                    }
                }
            }
            await fs.promises.rmdir(delPath);
            return true;
        }catch (err){}
        return false;
    }

    /**
     * Singleton: Creates a Lernstick Connector Service instance
     * @returns {LernstickConnectorService} Instance
     */
    constructor() {
        super();
        if (!!LernstickConnectorService.instance) {
            return LernstickConnectorService.instance;
        }
        LernstickConnectorService.instance  = this;
        this.eventHelper                    = EventHelper.getInstance();
        this.currentHardwareInfo            = null;
        this.currentExamination             = null;
        this.currentExaminationProcess      = null;
        this.documentsStorage               = path.join(__dirname, "documents");
        this.examinationDocuments           = [];
        this.examinationDocumentsMap        = new Map();
        this.documentsId                    = null;

        this.examinationFilePath            = path.join(config.services.LernstickConnector.examinationFileTargetDirectoryPath,
            config.services.LernstickConnector.examinationFilename);

        if (fs.existsSync(config.services.LernstickConnector.validationPublicKeyPath)){
            this.examPublicKey = fs.readFileSync(config.services.LernstickConnector.validationPublicKeyPath, "utf-8");
        }else{
            throw Error(`No validation public key found on ${config.services.LernstickConnector.validationPublicKeyPath}`);
        }

        let jsonSchema = null;
        if (fs.existsSync(config.services.LernstickConnector.validationJSONSchemaPath)){
            jsonSchema = JSON.parse(fs.readFileSync(config.services.LernstickConnector.validationJSONSchemaPath, "utf-8"));
        }else{
            throw Error(`No validation JSON schema found on ${config.services.LernstickConnector.validationJSONSchemaPath}`);
        }

        if (fs.existsSync(config.services.LernstickConnector.validationJSONSchemaPath)){
            jsonSchema = JSON.parse(fs.readFileSync(config.services.LernstickConnector.validationJSONSchemaPath, "utf-8"));
        }else{
            throw Error(`No validation JSON schema found on ${config.services.LernstickConnector.validationJSONSchemaPath}`);
        }

        if (!fs.existsSync(this.documentsStorage)) fs.mkdirSync(this.documentsStorage);

        this.jsonSchemaValidator            = ajvInstance.compile(jsonSchema);

        this.saveExaminationFile            = this.saveExaminationFile.bind(this);
        this.verifyExaminationFile          = this.verifyExaminationFile.bind(this);
        this.startExamination               = this.startExamination.bind(this);
        this.stopExamination                = this.stopExamination.bind(this);
        this.hashURL                        = this.hashURL.bind(this);
        this.getHardwareInfo                = this.getHardwareInfo.bind(this);
        this.getExamination                 = this.getExamination.bind(this);
        this.getExaminationDocuments        = this.getExaminationDocuments.bind(this);
        this.downloadExaminationDocument    = this.downloadExaminationDocument.bind(this);

        this._loadExamination               = this._loadExamination.bind(this);
        this._listenForHardwareInfo         = this._listenForHardwareInfo.bind(this);
        this._runCMDCommand                 = this._runCMDCommand.bind(this);
        this._updateExamination             = this._updateExamination.bind(this);
        this._verifyExaminationFileData     = this._verifyExaminationFileData.bind(this);
        this._killCurrentCMDCommand         = this._killCurrentCMDCommand.bind(this);
        this._exchangeExaminationData       = this._exchangeExaminationData.bind(this);
        this._handleExaminationProcessExit  = this._handleExaminationProcessExit.bind(this);

        this._listenForHardwareInfo();
        this._loadExamination();
        return this;
    }

    /**
     * Examination status
     * @return {Object}
     */
    getExamination(){
        return this.currentExamination ? {
            isRunning: this.currentExamination.isRunning,
            hasConfig: this.currentExamination.config !== null,
            name: this.currentExamination.config ? this.currentExamination.config.name : "",
            instanceId: this.currentExamination.config ? this.currentExamination.config.instance_id : "none",
            examType: this.currentExamination.config ? this.currentExamination.config.exam_type : "none",
            documents: this.examinationDocuments
        } : {
            isRunning: false,
            hasConfig: false,
            name: "",
            instanceId: "",
            examType: "none",
            documents: []
        }
    }

    /**
     * Hardware status
     * @return {Object|null}
     */
    getHardwareInfo(){
        return this.currentHardwareInfo;
    }

    /**
     * Get all downloaded examination documents
     * @return {Promise<Array<Object>>}
     */
    getExaminationDocuments(){
        return this.examinationDocuments;
    }

    /**
     * Start document download if the file exists
     * @param {String} id - Identification
     * @param {HTTPResponse} res - HTTP response
     */
    async downloadExaminationDocument(id, res){
        if (!this.examinationDocumentsMap.has(id)){
            res.sendStatus(404);
            return;
        }
        const file = this.examinationDocumentsMap.get(id);
        const target = path.join(this.documentsStorage, "extracted", id);
        if (fs.existsSync(target)){
            res.sendFile(target, {
                headers: {
                    "Content-Type": "application/pdf",
                    "Content-Disposition": `inline; filename="${file.name}"`
                }
            });
        }else{
            res.sendStatus(404);
        }

    }

    /**
     * Hashes given url with examination key
     * @param {String} url - Url to hash
     * @return {string}
     */
    hashURL(url){
        try {
            if (this.currentExamination
                && (typeof url === "string" || url instanceof String)
                && url.length > 0
                && this.currentExamination.config !== null
                && this.currentExamination.config["lms_examination_key"] !== "undefined"
                && this.currentExamination.config["lms_examination_key"].length > 0){
                return crypto
                    .createHash('sha256')
                    .update(`${url}${this.currentExamination.config["lms_examination_key"]}`)
                    .digest('hex');
            }
        }catch (err){}
        return "";
    }

    /**
     * Stop current examination process
     * @param {Boolean} force - If true, ignore if process is currently running
     * @return {Promise<Boolean>}
     */
    async stopExamination(force = false){
        if ((this.currentExaminationProcess && this.currentExamination.isRunning) || force){
            try {
                // Kill current process
                await this._killCurrentCMDCommand();
                this.currentExaminationProcess = null;

                // Cleanup examination data
                this.examinationDocuments = [];
                this.examinationDocumentsMap = new Map();
                this.documentsId = null;
                const targetZipPath = path.join(this.documentsStorage, "documents.zip");
                const extractedDirectory = path.join(this.documentsStorage, "extracted");
                const cacheFile = path.join(this.documentsStorage, "cache.json");
                if (fs.existsSync(targetZipPath)) await fs.promises.unlink(targetZipPath);
                if (fs.existsSync(cacheFile)) await fs.promises.unlink(cacheFile);
                await LernstickConnectorService.removeDirectory(extractedDirectory);
                if (fs.existsSync(this.examinationFilePath)) await fs.promises.unlink(this.examinationFilePath);

                this.currentExamination["config"] = null;
                this._updateExamination(this.currentExamination);

                return true;
            }catch (err){
                this._log().error(err);
            }
        }else{
            return true;
        }
        return false;
    }

    /**
     * Start examination process of lernstick
     * @return {Promise<String|null>}
     */
    async startExamination(){
        try {
            if (this.examinationFilePath
                && this.currentExamination
                && this.currentExamination.config !== null
                && fs.existsSync(this.examinationFilePath)
                && fs.existsSync(config.services.LernstickConnector.examinationStarterPath)){
                if (this.currentExaminationProcess){
                    this._log().info("Examination process is running");

                    // Kill current process
                    await this._killCurrentCMDCommand();
                    this.currentExaminationProcess = null;
                }

                const examinationDataStatus = await this._exchangeExaminationData();
                // If the examination type is local, exchange need to be successful
                if (this.currentExamination.config.exam_type === "local" && !examinationDataStatus) return null;

                const res = await this._runCMDCommand(config.services.LernstickConnector.examinationStarterPath,
                    config.services.LernstickConnector.examinationFileTargetDirectoryPath);
                return res;
            }
        }catch (err){
            this._log().error(err);
        }
        return null;
    }

    /**
     * Save examination file
     * @param {Object} examination
     * @param {Object} examinationConfig
     * @return {Promise<boolean>}
     */
    async saveExaminationFile(examination, examinationConfig){
        if (typeof examination["examinationFile"] === "undefined") return false;
        try {
            if (fs.existsSync(config.services.LernstickConnector.temporaryDirectory)
                && fs.existsSync(config.services.LernstickConnector.examinationFileTargetDirectoryPath)){
                await fs.promises.writeFile(this.examinationFilePath, examination.examinationFile, "base64");
                this._log().info("Saved new examination file");
                this._updateExamination({
                    isRunning: this.currentExamination ? this.currentExamination.isRunning : false,
                    config: examinationConfig
                });
                return true;
            }
        }catch (err){
            this._log().error(err);
        }
        return false;
    }

    /**
     * Read examination file
     * @return {Promise<Object|null>}
     */
    async readExaminationFile() {
        try {
            if (!fs.existsSync(this.examinationFilePath)) return null;

            // Extract files out zp archive
            const examinationZip = new streamZip.async({
                file: this.examinationFilePath,
                skipEntryNameValidation: true,
                storeEntries: true
            });
            const files = await examinationZip.entries();
            let examFileConf = null;
            let examFileSig = null;
            let examinationData = null;
            for (const file of Object.values(files)) {
                if (file.name === "exam.config") {
                    const tpmExamFileConf = (await examinationZip.entryData(file.name));
                    try {
                        examinationData = JSON.parse(tpmExamFileConf.toString("utf8"));
                        examFileConf = tpmExamFileConf;
                    }catch (err){
                        this._log().error("Invalid JSON string given");
                        this._log().error(err);
                    }
                }
                if (file.name === "exam.config.sig")
                    examFileSig = (await examinationZip.entryData(file.name));
            }
            await examinationZip.close();
            if (examFileConf === null || examFileSig === null) {
                this._log().error("Invalid examination file was given")
                return null;
            }

            // Verify examination file
            return this._verifyExaminationFileData(examinationData, examFileConf, examFileSig);
        }catch (err){
            this._log().error(err);
        }
        return null;
    }


    /**
     * Verify given examination
     * @param {Object} examination
     * @return {Promise<Object>}
     */
    async verifyExaminationFile(examination){
        if (typeof examination["examinationFile"] === "undefined") return null;
        const tmpFile = path.join(config.services.LernstickConnector.temporaryDirectory, `campla-${Date.now()}.txt`);
        try {
            if (fs.existsSync(config.services.LernstickConnector.temporaryDirectory)){
                await fs.promises.writeFile(tmpFile, examination.examinationFile, "base64");

                // Extract files out zp archive
                const examinationZip = new streamZip.async({
                    file: tmpFile,
                    skipEntryNameValidation: true,
                    storeEntries: true
                });
                const files = await examinationZip.entries();
                let examFileConf = null;
                let examFileSig = null;
                let examinationData = null;
                for (const file of Object.values(files)) {
                    if (file.name === "exam.config") {
                        const tpmExamFileConf = (await examinationZip.entryData(file.name));
                        try {
                            examinationData = JSON.parse(tpmExamFileConf.toString("utf8"));
                            examFileConf = tpmExamFileConf;
                        }catch (err){
                            this._log().error("Invalid JSON string given");
                            this._log().error(err);
                        }
                    }
                    if (file.name === "exam.config.sig")
                        examFileSig = (await examinationZip.entryData(file.name));
                }
                await examinationZip.close();
                if (examFileConf === null || examFileSig === null) {
                    this._log().error("Invalid examination file was given");
                    if (fs.existsSync(tmpFile)) await fs.promises.unlink(tmpFile);
                    return null;
                }

                if (fs.existsSync(tmpFile)) await fs.promises.unlink(tmpFile);

                // Verify examination file

                return this._verifyExaminationFileData(examinationData, examFileConf, examFileSig);
            }
        }catch (err){
            this._log().error(err);
            try {
                if (fs.existsSync(tmpFile)) await fs.promises.unlink(tmpFile);
            }catch (err){
                this._log().error(err);
            }
        }
        return null;
    }

    /**
     * Verify if the examination file is valid
     * @param {Object} examinationData - Examination configuration
     * @param {Buffer} examFileConf - Plain examination file content
     * @param {Buffer} examFileSig - Examination file signature
     * @return {null|Object}
     * @private
     */
    _verifyExaminationFileData(examinationData, examFileConf, examFileSig){
        if (crypto.verify("RSA-SHA256", examFileConf, this.examPublicKey, examFileSig)) {
            if (this.jsonSchemaValidator(examinationData)){
                if (examinationData.expiration > Date.now()){
                    this._log().info("Valid examination file given");
                    return examinationData;
                }else{
                    this._log().error("Examination file is expired");
                }
            }else{
                this._log().error("Invalid JSON object given");
            }
        }else{
            this._log().error("Invalid signature given");
        }
        return null;
    }

    /**
     * Listen for hardware info updates
     * @return {Promise<void>}
     * @private
     */
    async _listenForHardwareInfo(){
        let isFailed = false;
        setInterval(async () => {
            try {
                if (fs.existsSync(config.services.LernstickConnector.hardwareInfoPath)){
                    const oldHardwareStatus = this.currentHardwareInfo;
                    this.currentHardwareInfo = JSON.parse((await fs.promises.readFile(
                        config.services.LernstickConnector.hardwareInfoPath, "utf-8")));
                    if (oldHardwareStatus){
                        if (JSON.stringify(oldHardwareStatus) !== JSON.stringify(this.currentHardwareInfo))
                            this.eventHelper.fireEvent("update", "hardware", {
                                hardwareStatus: this.currentHardwareInfo
                            });
                    }else{
                        this.eventHelper.fireEvent("update", "hardware", {
                            hardwareStatus: this.currentHardwareInfo
                        });
                    }
                    isFailed = false;
                }
            }catch (err){
                if (!isFailed) this._log().error(err);
                isFailed = true;
            }
        }, 500);
    }

    /**
     * Load examination configuration from file
     * @return {Promise<void>}
     * @private
     */
    async _loadExamination(){
        this._updateExamination({
            isRunning: this.currentExaminationProcess !== null,
            config: (await this.readExaminationFile())
        });
        this.startExamination();
    }

    /**
     * Kill current examination process
     * @return {Promise<Boolean>}
     * @private
     */
    async _killCurrentCMDCommand(){
        return await new Promise((resolve) => {
            try {
                if (this.currentExaminationProcess === null) {
                    this._log().info("Examination process already killed");
                    this._updateExamination({ isRunning: false, config: this.currentExamination.config });
                    resolve(true);
                    return;
                }
                this.currentExaminationProcess.on('exit', () => {
                    clearTimeout(timeout);
                    this._log().info("Examination process killed successfully");
                    this._updateExamination({ isRunning: false, config: this.currentExamination.config });
                    resolve(true);
                })
                if (this.currentExaminationProcess) this.currentExaminationProcess.kill("SIGHUP");
                const timeout = setTimeout(() => {
                    if (this.currentExaminationProcess) this.currentExaminationProcess.kill("SIGKILL");
                }, 5000);
            }catch (err){
                resolve(false);
                this._log().error(err);
            }
        });
    }

    /**
     * Executes a given command line command
     * @param {String} command - Command line path
     * @param workingDirectory - Path to the working directory
     * @returns {Promise<String>}
     * @private
     */
    async _runCMDCommand(command, workingDirectory){
        this.currentExaminationProcess = spawn(command, [], {
            cwd: workingDirectory,
            shell: "/bin/bash"
        });
        this._log().info(`Examination process started`);
        this._updateExamination({ isRunning: true, config: this.currentExamination.config });
        this.currentExaminationProcess.on('exit', this._handleExaminationProcessExit);
        this.currentExaminationProcess.stdout.on('data', msg => this._log().info(`Exam start script (stdout): ${msg}`.trim()));
        this.currentExaminationProcess.stderr.on('data', msg => this._log().info(`Exam start script (stderr): ${msg}`.trim()));
        this.currentExaminationProcess.on('error', err =>
            this._log().error(`Examination process error: ${err}`));

    }

    /**
     * Handle examination process exit
     * @return {Promise<void>}
     * @private
     */
    async _handleExaminationProcessExit(){
        this._log().info(`Examination process exited`);
        this.currentExaminationProcess = null;
        this._updateExamination({ isRunning: false, config: this.currentExamination.config });
    }

    /**
     * Download examination documents and upload hardware statistics
     * @return {Promise<boolean>}
     * @private
     */
    async _exchangeExaminationData(){
        if (this.currentExamination.config === null) return false;

        // Target ZIP archive
        const targetZipPath = path.join(this.documentsStorage, "documents.zip");
        // Extracted document directory
        const extractedDirectory = path.join(this.documentsStorage, "extracted");
        // File list change
        const cacheFile = path.join(this.documentsStorage, "cache.json");

        if (this.documentsId === null || this.documentsId === this.currentExamination.config.instance_id){
            try {
                if (fs.existsSync(cacheFile)){
                    const examinationDocuments = JSON.parse((await fs.promises.readFile(cacheFile, "utf-8")));
                    if (examinationDocuments.length > 0){
                        const examinationDocumentsMap = new Map();
                        examinationDocuments.forEach(f => examinationDocumentsMap.set(f.id, f));
                        this.examinationDocuments = examinationDocuments;
                        this.examinationDocumentsMap = examinationDocumentsMap;
                    }
                    this.documentsId = this.currentExamination.config.instance_id;
                    return true;
                }
            }catch (err){
                this._log().error(err);
            }
        }else{
            try {
                this.examinationDocuments = [];
                this.examinationDocumentsMap = new Map();
                this.documentsId = null;
                this._updateExamination(this.currentExamination);

                if (fs.existsSync(targetZipPath)) await fs.promises.unlink(targetZipPath);
                if (fs.existsSync(cacheFile)) await fs.promises.unlink(cacheFile);
                await LernstickConnectorService.removeDirectory(extractedDirectory);
            }catch (err){
                this._log().error(err);
            }
        }

        const data = {
            auth: {
              id:  this.currentExamination.config.instance_id,
              token:  this.currentExamination.config.instance_token
            },
            hardware: this.currentHardwareInfo ? this.currentHardwareInfo.hardware : {}
        }
        try {
            const res = await fetch(`${config.services.LernstickConnector.camplaCoreApiUrl}`, {
                method: "POST",
                headers: {
                    "Content-Type": "application/json"
                },
                body: JSON.stringify(data)
            });

            // Request failed
            if (!res.ok) return false;

            const contentType = res.headers.get("Content-Type");
            if (contentType.toLowerCase() === "application/zip"){

                // Save ZIP file
                const buffer = await res.buffer();
                await fs.promises.writeFile(targetZipPath, buffer);

                const zip = new streamZip.async({ file: targetZipPath });

                // Read archive reset extracted directory
                const examinationDocuments = [];
                const examinationDocumentsMap = new Map();
                const submissionDirectory = path.join(extractedDirectory, "data");
                const entries = await zip.entries();
                await LernstickConnectorService.removeDirectory(extractedDirectory)
                await fs.promises.mkdir(extractedDirectory);
                await fs.promises.mkdir(submissionDirectory);

                // Extract files and list examination documents
                for (const entry of Object.values(entries)) {
                    const id = uuid.v4();
                    if (!entry.isDirectory){
                        const name = entry.name.replace(/\uFFFD/g, '');
                        await zip.extract(entry.name, path.join(extractedDirectory, id));
                        let document = {
                            id: id,
                            name: name,
                            submitted: false
                        };
                        if (name.indexOf("Data/") !== -1){
                            document["submitted"] = true;
                            document = {
                                id: id,
                                name: name.replace('Data/', ''),
                                submitted: true
                            };
                        }
                        examinationDocuments.push(document);
                        examinationDocumentsMap.set(id, document);
                    }
                }
                await zip.close();

                this.examinationDocuments = examinationDocuments;
                this.examinationDocumentsMap = examinationDocumentsMap;
                await fs.promises.writeFile(cacheFile, JSON.stringify(examinationDocuments));
                this.documentsId = this.currentExamination.config.instance_id;
                this._updateExamination(this.currentExamination);
            }else if (contentType.toLowerCase() === "application/json"){
                this.examinationDocuments = [];
                this.examinationDocumentsMap = new Map();
            }
            return true;
        }catch (err){
            this._log().error(err);
        }
        return false;
    }

    /**
     * Updates examination status
     * @param {Object} examination
     * @private
     */
    _updateExamination(examination){
        this.currentExamination = examination;
        this.eventHelper.fireEvent("update", "examination", {
            isRunning: examination.isRunning,
            hasConfig: examination.config !== null,
            name: examination.config ? examination.config.name : "",
            instanceId: examination.config ? examination.config.instance_id : "",
            examType: examination.config ? examination.config.exam_type : "none",
            documents: this.examinationDocuments
        });
    }

}

module.exports = LernstickConnectorService;
